@@ -16,4 +16,9 @@ import pkg_resources
 try:
     __version__ = pkg_resources.get_distribution('exposan').version
 except pkg_resources.DistributionNotFound:
-    __version__ = None
\ No newline at end of file
+    __version__ = None
+    
+import os
+es_path = os.path.dirname(__file__)
+del os
+    
\ No newline at end of file
