@@ -10,28 +10,25 @@ This module is developed by:
 This module is under the University of Illinois/NCSA Open Source License.
 Please refer to https://github.com/QSD-Group/EXPOsan/blob/main/LICENSE.txt
 for license details.
-
-TODOs:
-    [1] Recheck COD calculation acorss units once WasteStream is ready.
-
-Questions:
-    [1] WWTP power consumption very low, only 6.5/0.8 kWh per hour for
-        existing/alternative WWTP.
-
 '''
 
 
 # %%
 
+# Filter out warnings related to solid content
+import warnings
+warnings.filterwarnings('ignore')
+
 import numpy as np
 import biosteam as bst
 import qsdsan as qs
+from collections.abc import Iterable
 from sklearn.linear_model import LinearRegression as LR
 from qsdsan import sanunits as su
 from qsdsan import WasteStream, ImpactIndicator, ImpactItem, StreamImpactItem, SimpleTEA, LCA
+from exposan.bwaise import results_path
 from exposan.bwaise._cmps import cmps
-from exposan.bwaise._lca_data import load_lca_data
-
+from exposan.bwaise._lca_data import lca_data_kind, load_lca_data, _ImpactItem_LOADED
 
 # =============================================================================
 # Unit parameters
@@ -43,22 +40,28 @@ qs.set_thermo(cmps)
 bst.speed_up()
 
 household_size = 4
-get_household_size = lambda: household_size
 household_per_toilet = 4
-get_household_per_toilet = lambda: household_per_toilet
-get_toilet_user = lambda: get_household_size()*get_household_per_toilet()
+get_toilet_user = lambda: household_size * household_per_toilet
 
 # Number of people served by the existing plant (sysA and sysC)
 ppl_exist_sewer = 4e4
 ppl_exist_sludge = 416667
 # Number of people served by the alternative plant (sysB)
 ppl_alt = 5e4
-get_ppl = lambda kind: ppl_exist_sewer+ppl_exist_sludge if kind=='exist' else ppl_alt
+def get_ppl(kind):
+    if kind.lower() in ('exist', 'existing', 'sysa', 'sysc', 'a', 'c'):
+        return ppl_exist_sewer+ppl_exist_sludge
+    elif kind.lower() in ('alt', 'alternative', 'sysb', 'b'):
+        return ppl_alt
+    else:
+        raise ValueError('`kind` should be "exist" (for sysA and sysC)'
+                         f'or "alt" for sysB, not {kind}.')
+
 
 exchange_rate = 3700 # UGX per USD
 get_exchange_rate = lambda: exchange_rate
+
 discount_rate = 0.05
-get_discount_rate = lambda: discount_rate
 
 # Time take for full degradation, [yr]
 tau_deg = 2
@@ -71,30 +74,27 @@ def get_decay_k(tau_deg=2, log_deg=3):
     return k
 
 max_CH4_emission = 0.25
-get_max_CH4_emission = lambda: max_CH4_emission
 
 # Model for tanker truck cost based on capacity (m3)
 # price = a*capacity**b -> ln(price) = ln(a) + bln(capacity)
 UGX_price_dct = np.array((8e4, 12e4, 20e4, 25e4))
 capacities = np.array((3, 4.5, 8, 15))
-emptying_fee = 0.15
-get_emptying_fee = lambda: emptying_fee
+emptying_fee = 0.15 # additional emptying fee, fraction of base cost
 def get_tanker_truck_fee(capacity):
-    price_dct = UGX_price_dct*(1+get_emptying_fee())/get_exchange_rate()
+    price_dct = UGX_price_dct*(1+emptying_fee)/get_exchange_rate()
     ln_p = np.log(price_dct)
     ln_cap = np.log(capacities)
     model = LR().fit(ln_cap.reshape(-1,1), ln_p.reshape(-1,1))
-    [[predicted]] = model.predict(np.array((np.log(capacity))).reshape(1, -1)).tolist()
+    predicted = model.predict(np.array((np.log(capacity))).reshape(1, -1)).item()
     cost = np.exp(predicted)
     return cost
 
 # Flow rates for treatment plants
 sewer_flow = 2750 # m3/d
-get_sewer_flow = lambda: sewer_flow
 sludge_flow_exist = 500 # m3/d
 sludge_flow_alt = 60 # m3/d
 get_sludge_flow = lambda kind: \
-    sludge_flow_exist if kind=='exist' else sludge_flow_alt
+    sludge_flow_exist if kind.lower() in ('exist', 'sysa', 'sysc', 'a', 'c') else sludge_flow_alt
 
 # Nutrient loss during applciation
 app_loss = dict.fromkeys(('NH3', 'NonNH3', 'P', 'K', 'Mg', 'Ca'), 0.02)
@@ -102,10 +102,8 @@ app_loss['NH3'] = 0.05
 
 # Energetic content of the biogas
 biogas_energy = 803 # kJ/mol CH4
-get_biogas_energy = lambda: biogas_energy
 LPG_energy = 50 # MJ/kg
-get_LPG_energy = lambda: LPG_energy
-get_biogas_factor = lambda: get_biogas_energy()/cmps.CH4.MW/get_LPG_energy()
+get_biogas_factor = lambda: biogas_energy/cmps.CH4.MW/LPG_energy
 
 
 # =============================================================================
@@ -127,7 +125,7 @@ price_dct = {
     }
 
 GWP_dct = {
-    'Electricity': 0.15,
+    'Electricity': 0.1135,
     'CH4': 28,
     'N2O': 265,
     'N': -5.4,
@@ -136,16 +134,14 @@ GWP_dct = {
     'Biogas': -3*get_biogas_factor()
     }
 
-items = ImpactItem.get_all_items()
-
-if not items.get('Excavation'): # prevent from reloading
+if not ImpactItem.get_item('Excavation'): # prevent from reloading
     load_lca_data('original')
 
 GWP = ImpactIndicator.get_indicator('GWP')
 
 bst.PowerUtility.price = price_dct['Electricity']
-items['Concrete'].price = price_dct['Concrete']
-items['Steel'].price = price_dct['Steel']
+ImpactItem.get_item('Concrete').price = price_dct['Concrete']
+ImpactItem.get_item('Steel').price = price_dct['Steel']
 
 # =============================================================================
 # Universal units and functions
@@ -165,54 +161,89 @@ def batch_create_stream_items(kind):
         H_factor = {'GW2ECO': 0.0000000028, 'GW2HH': 0.000000928, 'OD2HH': 0.000531}
         I_factor = {'GW2ECO': 0.000000025, 'GW2HH': 0.0000125, 'OD2HH': 0.000237}
 
-        StreamImpactItem(ID=f'CH4_item',
-                         E_ClimateChangeEcosystems=E_factor['GW2ECO']*4.8,
-                         E_ClimateChangeHumanHealth=E_factor['GW2HH']*4.8,
-                         H_ClimateChangeEcosystems=H_factor['GW2ECO']*34,
-                         H_ClimateChangeHumanHealth=H_factor['GW2HH']*34,
-                         I_ClimateChangeEcosystems=I_factor['GW2ECO']*84,
-                         I_ClimateChangeHumanHealth=I_factor['GW2HH']*84
+        StreamImpactItem(ID='CH4_item',
+                         E_EcosystemQuality_Total=E_factor['GW2ECO']*4.8,
+                         E_HumanHealth_Total=E_factor['GW2HH']*4.8,
+                         H_EcosystemQuality_Total=H_factor['GW2ECO']*34,
+                         H_HumanHealth_Total=H_factor['GW2HH']*34,
+                         I_EcosystemQuality_Total=I_factor['GW2ECO']*84,
+                         I_HumanHealth_Total=I_factor['GW2HH']*84
                          )
-        StreamImpactItem(ID=f'N2O_item',
-                         E_ClimateChangeEcosystems=E_factor['GW2ECO']*78.8,
-                         E_ClimateChangeHumanHealth=E_factor['GW2HH']*78.8,
-                         H_ClimateChangeEcosystems=H_factor['GW2ECO']*298,
-                         H_ClimateChangeHumanHealth=H_factor['GW2HH']*298,
-                         I_ClimateChangeEcosystems=I_factor['GW2ECO']*264,
-                         I_ClimateChangeHumanHealth=I_factor['GW2HH']*264,
-                         E_OzoneDepletion=E_factor['OD2HH']*0.017,
-                         H_OzoneDepletion=H_factor['OD2HH']*0.011,
-                         I_OzoneDepletion=I_factor['OD2HH']*0.007
+        StreamImpactItem(ID='N2O_item',
+                         E_EcosystemQuality_Total=E_factor['GW2ECO']*78.8,
+                         # From climate change + ozone depletion
+                         E_HumanHealth_Total=\
+                             E_factor['GW2HH']*78.8+E_factor['OD2HH']*0.017,
+                         H_EcosystemQuality_Total=H_factor['GW2ECO']*298,
+                         H_HumanHealth_Total=\
+                             H_factor['GW2HH']*298+H_factor['OD2HH']*0.011,
+                         I_EcosystemQuality_Total=I_factor['GW2ECO']*264,
+                         I_HumanHealth_Total=\
+                             I_factor['GW2HH']*264+I_factor['OD2HH']*0.007
                          )
     else:
         raise ValueError(f'`kind` can only be "original" or "new", not "{kind}".')
 
-batch_create_stream_items(kind='original')
+    global _ImpactItem_LOADED
+    _ImpactItem_LOADED = True
+
 
 def batch_create_streams(prefix):
+    if not _ImpactItem_LOADED:
+        batch_create_stream_items(kind=lca_data_kind)
+
     stream_dct = {}
-    stream_dct['CH4'] = WasteStream(f'{prefix}_CH4', phase='g',
-                                    stream_impact_item=items['CH4_item'].copy(set_as_source=True))
-    stream_dct['N2O'] = WasteStream(f'{prefix}_N2O', phase='g',
-                                    stream_impact_item=items['N2O_item'].copy(set_as_source=True))
+    item = ImpactItem.get_item('CH4_item').copy(f'{prefix}_CH4_item', set_as_source=True)
+    stream_dct['CH4'] = WasteStream(f'{prefix}_CH4', phase='g', stream_impact_item=item)
+    # CH4.stream_impact_item = ImpactItem.get_item('CH4_item').copy(stream=CH4, set_as_source=True)
+
+    item = ImpactItem.get_item('N2O_item').copy(f'{prefix}_N2O_item', set_as_source=True)
+    stream_dct['N2O'] = WasteStream(f'{prefix}_N2O', phase='g', stream_impact_item=item)
+    # N2O.stream_impact_item = ImpactItem.get_item('N2O_item').copy(stream=N2O, set_as_source=True)
+
+    item = ImpactItem.get_item('N_item').copy(f'{prefix}_liq_N_item', set_as_source=True)
     stream_dct['liq_N'] = WasteStream(f'{prefix}_liq_N', phase='l', price=price_dct['N'],
-                                      stream_impact_item=items['N_item'].copy(set_as_source=True))
+                                      stream_impact_item=item)
+    # liq_N.stream_impact_item = ImpactItem.get_item('N_item').copy(stream=liq_N, set_as_source=True)
+
+    item = ImpactItem.get_item('N_item').copy(f'{prefix}_sol_N_item', set_as_source=True)
     stream_dct['sol_N'] = WasteStream(f'{prefix}_sol_N', phase='l', price=price_dct['N'],
-                                      stream_impact_item=items['N_item'].copy(set_as_source=True))
+                                              stream_impact_item=item)
+    # sol_N.stream_impact_item = ImpactItem.get_item('N_item').copy(stream=sol_N, set_as_source=True)
+
+    item = ImpactItem.get_item('P_item').copy(f'{prefix}_liq_P_item', set_as_source=True)
     stream_dct['liq_P'] = WasteStream(f'{prefix}_liq_P', phase='l', price=price_dct['P'],
-                                      stream_impact_item=items['P_item'].copy(set_as_source=True))
+                                      stream_impact_item=item)
+    # liq_P.stream_impact_item=ImpactItem.get_item('P_item').copy(stream=liq_P, set_as_source=True)
+
+    item = ImpactItem.get_item('P_item').copy(f'{prefix}_sol_P_item', set_as_source=True)
     stream_dct['sol_P'] = WasteStream(f'{prefix}_sol_P', phase='l', price=price_dct['P'],
-                                      stream_impact_item=items['P_item'].copy(set_as_source=True))
+                                      stream_impact_item=item)
+    # sol_P.stream_impact_item=ImpactItem.get_item('P_item').copy(stream=sol_P, set_as_source=True)
+
+    item = ImpactItem.get_item('K_item').copy(f'{prefix}_liq_K_item', set_as_source=True)
     stream_dct['liq_K'] = WasteStream(f'{prefix}_liq_K', phase='l', price=price_dct['K'],
-                                      stream_impact_item=items['K_item'].copy(set_as_source=True))
+                                      stream_impact_item=item)
+    # liq_K.stream_impact_item = ImpactItem.get_item('K_item').copy(stream=liq_K, set_as_source=True)
+
+    item = ImpactItem.get_item('K_item').copy(f'{prefix}_sol_K_item', set_as_source=True)
     stream_dct['sol_K'] = WasteStream(f'{prefix}_sol_K', phase='l', price=price_dct['K'],
-                                      stream_impact_item=items['K_item'].copy(set_as_source=True))
+                                      stream_impact_item=item)
     return stream_dct
 
+def update_toilet_param(unit, kind):
+    unit.N_user = get_toilet_user()
+    unit.N_toilet = get_ppl(kind)/get_toilet_user()
+    unit._run()
+
+def update_lagoon_flow_rate(unit):
+    unit.flow_rate = sewer_flow + get_sludge_flow('exist')
+    unit._run()
+
 def add_fugitive_items(unit, item_ID):
     unit._run()
     for i in unit.ins:
-        i.stream_impact_item = items[item_ID].copy(set_as_source=True)
+        i.stream_impact_item = ImpactItem.get_item(item_ID).copy(set_as_source=True)
 
 # Costs of WWTP units have been considered in the lumped unit
 def clear_unit_costs(sys):
@@ -220,6 +251,7 @@ def clear_unit_costs(sys):
         if isinstance(i, su.LumpedCost):
             continue
         i.purchase_costs.clear()
+        i.installed_costs.clear()
 
 
 def adjust_NH3_loss(unit):
@@ -251,7 +283,9 @@ A2 = su.PitLatrine('A2', ins=(A1-0, A1-1,
                    OPEX_over_CAPEX=0.05,
                    decay_k_COD=get_decay_k(tau_deg, log_deg),
                    decay_k_N=get_decay_k(tau_deg, log_deg),
-                   max_CH4_emission=get_max_CH4_emission())
+                   max_CH4_emission=max_CH4_emission
+                   )
+A2.specification = lambda: update_toilet_param(A2, 'exist')
 
 ##################### Conveyance #####################
 A3 = su.Trucking('A3', ins=A2-0, outs=('transported', 'conveyance_loss'),
@@ -275,32 +309,41 @@ A4 = su.LumpedCost('A4', ins=A3-0, cost_item_name='Lumped WWTP',
                    CAPEX=18606700, power=57120/(365*24), lifetime=8)
 A4.line = 'Lumped WWTP cost'
 get_A4_lifetime = lambda: A4.lifetime
+def update_A_t0():
+    A5.t0 = A2.emptying_period
+    A6.t0 = A5.t0 # A5.tau is for the solids
+    A7.t0 = A6.t0 + A6.tau/365
+    A8.t0 = A5.t0 + A5.tau/365
+A4.specification = update_A_t0
+A4.run_after_specification = True
 
 A5 = su.SedimentationTank('A5', ins=A4-0,
                           outs=('liq', 'sol', 'A5_CH4', 'A5_N2O'),
                           decay_k_COD=get_decay_k(tau_deg, log_deg),
                           decay_k_N=get_decay_k(tau_deg, log_deg),
-                          max_CH4_emission=get_max_CH4_emission())
+                          max_CH4_emission=max_CH4_emission)
 
 A6 = su.Lagoon('A6', ins=A5-0, outs=('anaerobic_treated', 'A6_CH4', 'A6_N2O'),
                design_type='anaerobic',
-               flow_rate=get_sewer_flow()+get_sludge_flow('exist'),
+               flow_rate=sewer_flow+get_sludge_flow('exist'),
                decay_k_N=get_decay_k(tau_deg, log_deg),
-               max_CH4_emission=get_max_CH4_emission())
+               max_CH4_emission=max_CH4_emission)
+A6.specification = lambda: update_lagoon_flow_rate(A6)
 
 A7 = su.Lagoon('A7', ins=A6-0, outs=('facultative_treated', 'A7_CH4', 'A7_N2O'),
                design_type='facultative',
-               flow_rate=get_sewer_flow()+get_sludge_flow('exist'),
+               flow_rate=sewer_flow+get_sludge_flow('exist'),
                decay_k_N=get_decay_k(tau_deg, log_deg),
-               max_CH4_emission=get_max_CH4_emission(),
+               max_CH4_emission=max_CH4_emission,
                if_N2O_emission=True)
+A7.specification = lambda: update_lagoon_flow_rate(A7)
 
 A8 = su.DryingBed('A8', ins=A5-1, outs=('dried_sludge', 'evaporated',
                                         'A8_CH4', 'A8_N2O'),
                   design_type='unplanted',
                   decay_k_COD=get_decay_k(tau_deg, log_deg),
                   decay_k_N=get_decay_k(tau_deg, log_deg),
-                  max_CH4_emission=get_max_CH4_emission())
+                  max_CH4_emission=max_CH4_emission)
 treatA = bst.System('treatA', path=(A4, A5, A6, A7, A8))
 A8._cost = lambda: clear_unit_costs(treatA)
 
@@ -330,14 +373,16 @@ A13 = su.ComponentSplitter('A13', ins=A9-0,
 ############### Simulation, TEA, and LCA ###############
 sysA = bst.System('sysA', path=(A1, A2, A3, treatA, A9, A10, A11, A12, A13))
 
-teaA = SimpleTEA(system=sysA, discount_rate=get_discount_rate(), start_year=2018,
+exist_staff_num = 12
+get_annual_labor = lambda: exist_staff_num*3e6*12/get_exchange_rate()
+teaA = SimpleTEA(system=sysA, discount_rate=discount_rate, start_year=2018,
                  lifetime=get_A4_lifetime(), uptime_ratio=1, lang_factor=None,
-                 annual_maintenance=0, annual_labor=12*3e6*12/get_exchange_rate(),
-                 construction_schedule=None)
+                 annual_maintenance=0, annual_labor=get_annual_labor())
 
 lcaA = LCA(system=sysA, lifetime=get_A4_lifetime(), lifetime_unit='yr', uptime_ratio=1,
-            # Assuming all additional WWTP OPEX from electricity
-            E_item=lambda: A4.power_utility.rate*(365*24)*8)
+           annualize_construction=True,
+           # Assuming all additional WWTP OPEX from electricity
+           E_item=lambda: A4.power_utility.rate*(365*24)*get_A4_lifetime())
 
 
 # %%
@@ -349,8 +394,9 @@ lcaA = LCA(system=sysA, lifetime=get_A4_lifetime(), lifetime_unit='yr', uptime_r
 flowsheetB = bst.Flowsheet('sysB')
 bst.main_flowsheet.set_flowsheet(flowsheetB)
 streamsB = batch_create_streams('B')
+B_biogas_item = ImpactItem.get_item('Biogas_item').copy('B_biogas_item', set_as_source=True)
 streamsB['biogas'] = WasteStream('B_biogas', phase='g', price=price_dct['Biogas'],
-                                 stream_impact_item=items['Biogas_item'].copy(set_as_source=True))
+                                 stream_impact_item=B_biogas_item)
 
 #################### Human Inputs ####################
 B1 = su.Excretion('B1', outs=('urine', 'feces'))
@@ -364,7 +410,8 @@ B2 = su.PitLatrine('B2', ins=(B1-0, B1-1,
                    OPEX_over_CAPEX=0.05,
                    decay_k_COD=get_decay_k(tau_deg, log_deg),
                    decay_k_N=get_decay_k(tau_deg, log_deg),
-                   max_CH4_emission=get_max_CH4_emission())
+                   max_CH4_emission=max_CH4_emission)
+B2.specification = lambda: update_toilet_param(B2, 'alt')
 
 ##################### Conveyance #####################
 B3 = su.Trucking('B3', ins=B2-0, outs=('transported', 'conveyance_loss'),
@@ -387,25 +434,34 @@ B4 = su.LumpedCost('B4', ins=B3-0, cost_item_name='Lumped WWTP',
                    CAPEX=337140, power=6854/(365*24), lifetime=10)
 B4.line = 'Lumped WWTP cost'
 get_B4_lifetime = lambda: B4.lifetime
+def update_B_t0():
+    B5.t0 = B2.emptying_period
+    B8.t0 = B7.t0 = B5.t0 + B5.tau/365
+B4.specification = update_B_t0
+B4.run_after_specification = True
 
 B5 = su.AnaerobicBaffledReactor('B5', ins=B4-0, outs=('ABR_treated', 'biogas',
                                                       'B5_CH4', 'B5_N2O'),
                                 decay_k_COD=get_decay_k(tau_deg, log_deg),
-                                max_CH4_emission=get_max_CH4_emission())
+                                max_CH4_emission=max_CH4_emission)
+def update_B5_gravel_density():
+    B5.gravel_density = B2.density_dct['Gravel']
+B5.specification = update_B5_gravel_density
+B5.run_after_specification = True
 
 B6 = su.SludgeSeparator('B6', ins=B5-0, outs=('liq', 'sol'))
 
 B7 = su.LiquidTreatmentBed('B7', ins=B6-0, outs=('liquid_bed_treated', 'B7_CH4', 'B7_N2O'),
                            decay_k_COD=get_decay_k(tau_deg, log_deg),
                            decay_k_N=get_decay_k(tau_deg, log_deg),
-                           max_CH4_emission=get_max_CH4_emission())
+                           max_CH4_emission=max_CH4_emission)
 
 B8 = su.DryingBed('B8', ins=B6-1, outs=('dried_sludge', 'evaporated',
                                         'B8_CH4', 'B8_N2O'),
                   design_type='planted',
                   decay_k_COD=get_decay_k(tau_deg, log_deg),
                   decay_k_N=get_decay_k(tau_deg, log_deg),
-                  max_CH4_emission=get_max_CH4_emission())
+                  max_CH4_emission=max_CH4_emission)
 
 treatB = bst.System('treatB', path=(B4, B5, B6, B7, B8))
 B8._cost = lambda: clear_unit_costs(treatB)
@@ -442,20 +498,25 @@ B15 = su.Mixer('B15', ins=(B14-0, B14-2), outs=streamsB['biogas'])
 ############### Simulation, TEA, and LCA ###############
 sysB = bst.System('sysB', path=(B1, B2, B3, treatB, B9, B10, B11, B12, B13, B14, B15))
 
+skilled_num = 5
+get_skilled_num = lambda: skilled_num
+skilled_salary = 5e6 # UGX/month
+get_tot_skilled_salary = lambda: skilled_salary*get_skilled_num()
+
 unskilled_num = 5
 get_unskilled_num = lambda: unskilled_num
 unskilled_salary = 75e4 # UGX/month
-get_unskilled_salary = lambda: unskilled_salary*get_unskilled_num()
-get_alt_salary = lambda: (5*5e6+get_unskilled_salary())*12/get_exchange_rate()
+get_tot_unskilled_salary = lambda: unskilled_salary*get_unskilled_num()
+get_alt_salary = lambda: (get_tot_skilled_salary()+get_tot_unskilled_salary())*12/get_exchange_rate()
 
-teaB = SimpleTEA(system=sysB, discount_rate=get_discount_rate(), start_year=2018,
+teaB = SimpleTEA(system=sysB, discount_rate=discount_rate, start_year=2018,
                   lifetime=get_B4_lifetime(), uptime_ratio=1, lang_factor=None,
-                  annual_maintenance=0, annual_labor=get_alt_salary(),
-                  construction_schedule=None)
+                  annual_maintenance=0, annual_labor=get_alt_salary())
 
 lcaB = LCA(system=sysB, lifetime=get_B4_lifetime(), lifetime_unit='yr', uptime_ratio=1,
+           annualize_construction=True,
            # Assuming all additional WWTP OPEX from electricity
-           E_item=lambda: B4.power_utility.rate*(365*24)*10)
+           E_item=lambda: B4.power_utility.rate*(365*24)*get_B4_lifetime())
 
 
 # %%
@@ -481,18 +542,19 @@ C2 = su.UDDT('C2', ins=(C1-0, C1-1,
              OPEX_over_CAPEX=0.1,
              decay_k_COD=get_decay_k(tau_deg, log_deg),
              decay_k_N=get_decay_k(tau_deg, log_deg),
-             max_CH4_emission=get_max_CH4_emission())
+             max_CH4_emission=max_CH4_emission)
+C2.specification = lambda: update_toilet_param(C2, 'exist')
 
 ##################### Conveyance #####################
-# Liquid waste
 handcart_fee = 0.01 # USD/cap/d
-get_handcart_fee = lambda: handcart_fee
 truck_fee = 23e3 # UGX/m3
-get_truck_fee = lambda: truck_fee
 
+# Handcart fee is for both liquid/solid
 get_handcart_and_truck_fee = \
-    lambda vol, ppl: get_truck_fee()/get_exchange_rate()*vol \
-        + get_handcart_fee()*ppl*C2.collection_period
+    lambda vol, ppl, include_fee: truck_fee/get_exchange_rate()*vol \
+        + int(include_fee)*handcart_fee*ppl*C2.collection_period
+
+# Liquid waste
 C3 = su.Trucking('C3', ins=C2-0, outs=('transported_l', 'loss_l'),
                  load_type='mass', distance=5, distance_unit='km',
                  loss_ratio=0.02)
@@ -511,8 +573,8 @@ def update_C3_C4_param():
     truck4.load = C4.F_mass_in * hr / C2.N_toilet
     rho3 = C3.F_mass_in/C3.F_vol_in
     rho4 = C4.F_mass_in/C4.F_vol_in
-    C3.fee = get_handcart_and_truck_fee(truck3.load/rho3, ppl)
-    C4.fee = get_handcart_and_truck_fee(truck4.load/rho4, ppl)
+    C3.fee = get_handcart_and_truck_fee(truck3.load/rho3, ppl, True)
+    C4.fee = get_handcart_and_truck_fee(truck4.load/rho4, ppl, False)
     C3._design()
     C4._design()
 C4.specification = update_C3_C4_param
@@ -522,25 +584,32 @@ C5 = su.LumpedCost('C5', ins=(C3-0, C4-0),
                    cost_item_name='Lumped WWTP',
                    CAPEX=18606700, power=57120/(365*24), lifetime=8)
 get_C5_lifetime = lambda: C5.lifetime
+def update_C_t0():
+    C8.t0 = C6.t0 = C2.collection_period / 365
+    C7.t0 = C6.t0 + C6.tau/365
+C5.specification = update_C_t0
+C5.run_after_specification = True
 
 C6 = su.Lagoon('C6', ins=C5-0, outs=('anaerobic_treated', 'C6_CH4', 'C6_N2O'),
                design_type='anaerobic',
-               flow_rate=get_sewer_flow()+get_sludge_flow('exist'),
+               flow_rate=sewer_flow+get_sludge_flow('exist'),
                decay_k_N=get_decay_k(tau_deg, log_deg),
-               max_CH4_emission=get_max_CH4_emission())
+               max_CH4_emission=max_CH4_emission)
+C6.specification = lambda: update_lagoon_flow_rate(C6)
 
 C7 = su.Lagoon('C7', ins=C6-0, outs=('facultative_treated', 'C7_CH4', 'C7_N2O'),
                design_type='facultative',
-               flow_rate=get_sewer_flow()+get_sludge_flow('exist'),
+               flow_rate=sewer_flow+get_sludge_flow('exist'),
                decay_k_N=get_decay_k(tau_deg, log_deg),
-               max_CH4_emission=get_max_CH4_emission(),
+               max_CH4_emission=max_CH4_emission,
                if_N2O_emission=True)
+C7.specification = lambda: update_lagoon_flow_rate(C7)
 
 C8 = su.DryingBed('C8', ins=C5-1, outs=('dried_sludge', 'evaporated', 'C8_CH4', 'C8_N2O'),
                  design_type='unplanted',
                  decay_k_COD=get_decay_k(tau_deg, log_deg),
                  decay_k_N=get_decay_k(tau_deg, log_deg),
-                 max_CH4_emission=get_max_CH4_emission())
+                 max_CH4_emission=max_CH4_emission)
 treatC = bst.System('treatC', path=(C5, C6, C7, C8))
 C8._cost = lambda: clear_unit_costs(treatC)
 
@@ -570,14 +639,14 @@ C13 = su.ComponentSplitter('C13', ins=C9-0,
 ############### Simulation, TEA, and LCA ###############
 sysC = bst.System('sysC', path=(C1, C2, C3, C4, treatC, C9, C10, C11, C12, C13))
 
-teaC = SimpleTEA(system=sysC, discount_rate=get_discount_rate(), start_year=2018,
+teaC = SimpleTEA(system=sysC, discount_rate=discount_rate, start_year=2018,
                  lifetime=get_C5_lifetime(), uptime_ratio=1, lang_factor=None,
-                 annual_maintenance=0, annual_labor=12*3e6*12/get_exchange_rate(),
-                 construction_schedule=None)
+                 annual_maintenance=0, annual_labor=12*3e6*12/get_exchange_rate())
 
 lcaC = LCA(system=sysC, lifetime=get_C5_lifetime(), lifetime_unit='yr', uptime_ratio=1,
+           annualize_construction=True,
            # Assuming all additional WWTP OPEX from electricity
-           E_item=lambda: C5.power_utility.rate*(365*24)*8)
+           E_item=lambda: C5.power_utility.rate*(365*24)*get_C5_lifetime())
 
 
 # %%
@@ -598,15 +667,29 @@ def update_lca_data(kind):
         "new" loads the data for ReCiPe and TRACI
         (ecoinvent 3.7.1, at the point of substitution).
     '''
+    global lca_data_kind
 
-    load_lca_data(kind)
-    batch_create_stream_items(kind)
+    if lca_data_kind != kind:
+        load_lca_data(kind)
+        batch_create_stream_items(kind)
 
-    for v in items['Biogas_item'].CFs.values():
-        v *= get_biogas_factor()
+        for lca in (lcaA, lcaB, lcaC):
+            for i in lca.lca_streams:
+                # To refresh the impact items
+                source_ID = i.stream_impact_item.source.ID
+                i.stream_impact_item.source = ImpactItem.get_item(source_ID)
+
+        Biogas_CFs = ImpactItem.get_item('Biogas_item').CFs
+        for k, v in Biogas_CFs.items():
+            Biogas_CFs[k] = v * get_biogas_factor()
+
+        for i in sysA, sysB, sysC:
+            i.simulate()
+
+        lca_data_kind = kind
 
 
-def get_total_inputs(unit):
+def get_total_inputs(unit, multiplier=1):
     if len(unit.ins) == 0: # Excretion units do not have ins
         ins = unit.outs
     else:
@@ -617,38 +700,29 @@ def get_total_inputs(unit):
     inputs['NH3'] = sum(i.imass['NH3'] for i in ins)
     inputs['P'] = sum(i.TP*i.F_vol/1e3 for i in ins)
     inputs['K'] = sum(i.TK*i.F_vol/1e3 for i in ins)
-    hr = 365 * 24
     for i, j in inputs.items():
-        inputs[i] = j * hr
+        inputs[i] = j * multiplier
     return inputs
 
-def get_recovery(ins=None, outs=None, hr=365*24, ppl=1, if_relative=True):
-    try: iter(outs)
-    except: outs = (outs,)
-    non_g = tuple(i for i in outs if i.phase != 'g')
+
+def get_recovery(ins, outs, multiplier=1):
+    if not isinstance(outs, Iterable):
+        outs = (outs,)
+
+    non_g = tuple(i for i in outs if (i.phase != 'g' and 'los' not in i.ID))
     recovery = {}
     recovery['COD'] = sum(i.COD*i.F_vol/1e3 for i in non_g)
     recovery['N'] = sum(i.TN*i.F_vol/1e3 for i in non_g)
     recovery['NH3'] = sum(i.imass['NH3'] for i in non_g)
     recovery['P'] = sum(i.TP*i.F_vol/1e3 for i in non_g)
     recovery['K'] = sum(i.TK*i.F_vol/1e3 for i in non_g)
+
     for i, j in recovery.items():
-        if if_relative:
-            inputs = get_total_inputs(ins)
-            recovery[i] /= inputs[i]/hr * ppl
-        else:
-            recovery[i] /= 1/hr * ppl
+        inputs = get_total_inputs(ins, multiplier)
+        recovery[i] /= inputs[i]
+
     return recovery
 
-def get_stream_emissions(streams=None, hr=365*24, ppl=1):
-    try: iter(streams)
-    except: streams = (streams,)
-    emission = {}
-    factor = hr / ppl
-    for i in streams:
-        if not i.stream_impact_item: continue
-        emission[f'{i.ID}'] = i.F_mass*i.stream_impact_item.CFs['GlobalWarming']*factor
-    return emission
 
 sys_dct = {
     'ppl': dict(sysA=get_ppl('exist'), sysB=get_ppl('alt'), sysC=get_ppl('exist')),
@@ -662,57 +736,41 @@ sys_dct = {
     'cache': dict(sysA={}, sysB={}, sysC={}),
     }
 
+
 def cache_recoveries(sys):
-    total_COD = get_total_inputs(sys_dct['input_unit'][sys.ID])['COD']
-    ppl = sys_dct['ppl'][sys.ID]
+    sys_dct['ppl'][sys.ID] = ppl = get_ppl('alt') if sys.ID=='sysB' else get_ppl('exist')
+    total_COD = get_total_inputs(sys_dct['input_unit'][sys.ID], ppl)['COD']
 
     if sys_dct['gas_unit'][sys.ID]:
         gas_mol = sys_dct['gas_unit'][sys.ID].outs[0].imol['CH4']
-        gas_COD = gas_mol*1e3*get_biogas_energy()*365*24/14e3/ppl/total_COD
+        gas_COD = gas_mol*1e3*biogas_energy/14e3/total_COD
     else:
         gas_COD = 0
 
-    cache = {
+    sys_dct['cache'][sys.ID] = cache = {
         'liq': get_recovery(ins=sys_dct['input_unit'][sys.ID],
                             outs=sys_dct['liq_unit'][sys.ID].ins,
-                            ppl=ppl),
+                            multiplier=ppl),
         'sol': get_recovery(ins=sys_dct['input_unit'][sys.ID],
                             outs=sys_dct['sol_unit'][sys.ID].ins,
-                            ppl=ppl),
+                            multiplier=ppl),
         'gas': dict(COD=gas_COD, N=0, P=0, K=0)
         }
     return cache
 
-def update_cache(sys):
-    last_u = sys.path[-1]
-    last_u._run()
-    sys_dct['cache'][sys.ID] = cache_recoveries(sys)
 
-A13.specification = lambda: update_cache(sysA)
-B15.specification = lambda: update_cache(sysB)
-C13.specification = lambda: update_cache(sysC)
+sysA._set_facilities([*sysA.facilities, lambda: cache_recoveries(sysA)])
+sysB._set_facilities([*sysB.facilities, lambda: cache_recoveries(sysB)])
+sysC._set_facilities([*sysC.facilities, lambda: cache_recoveries(sysC)])
 
 
-def get_summarizing_fuctions():
+def get_summarizing_functions(system):
     func_dct = {}
-    func_dct['get_annual_cost'] = lambda tea, ppl: tea.EAC/ppl
-    func_dct['get_annual_CAPEX'] = lambda tea, ppl: tea.annualized_CAPEX/ppl
+    func_dct['get_annual_net_cost'] = lambda tea, ppl: (tea.annualized_equipment_cost-tea.net_earnings)/ppl
+    func_dct['get_annual_CAPEX'] = lambda tea, ppl: tea.annualized_equipment_cost/ppl
     func_dct['get_annual_OPEX'] = lambda tea, ppl: tea.AOC/ppl
-    ind = 'GlobalWarming'
-    func_dct['get_annual_GWP'] = \
-        lambda lca, ppl: lca.total_impacts[ind]/lca.lifetime/ppl
-    func_dct['get_constr_GWP'] = \
-        lambda lca, ppl: lca.total_construction_impacts[ind]/lca.lifetime/ppl
-    func_dct['get_trans_GWP'] = \
-        lambda lca, ppl: lca.total_transportation_impacts[ind]/lca.lifetime/ppl
-    func_dct['get_direct_emission_GWP'] = \
-        lambda lca, ppl: lca.get_stream_impacts(stream_items=lca.stream_inventory, kind='direct_emission')[ind] \
-            /lca.lifetime/ppl
-    func_dct['get_offset_GWP'] = \
-        lambda lca, ppl: lca.get_stream_impacts(stream_items=lca.stream_inventory, kind='offset')[ind] \
-            /lca.lifetime/ppl
-    func_dct['get_other_GWP'] = \
-        lambda lca, ppl: lca.total_other_impacts[ind]/lca.lifetime/ppl
+    func_dct['get_annual_sales'] = lambda tea, ppl: tea.sales/ppl
+
     for i in ('COD', 'N', 'P', 'K'):
         func_dct[f'get_liq_{i}_recovery'] = \
             lambda sys, i: sys_dct['cache'][sys.ID]['liq'][i]
@@ -725,55 +783,75 @@ def get_summarizing_fuctions():
                 sys_dct['cache'][sys.ID]['liq'][i] + \
                 sys_dct['cache'][sys.ID]['sol'][i] + \
                 sys_dct['cache'][sys.ID]['gas'][i]
+
     return func_dct
 
 
 def print_summaries(systems):
-    try: iter(systems)
-    except: systems = (systems, )
-    func = get_summarizing_fuctions()
+    if not isinstance(systems, Iterable):
+        systems = (systems, )
+
     for sys in systems:
+        func = get_summarizing_functions(sys)
         sys.simulate()
         ppl = sys_dct['ppl'][sys.ID]
         print(f'\n---------- Summary for {sys} ----------\n')
-        tea = sys_dct['TEA'][sys.ID]
-        tea.show()
-        print('\n')
-        lca = sys_dct['LCA'][sys.ID]
-        lca.show()
-
-        unit = f'{currency}/cap/yr'
-        print(f'\nNet cost: {func["get_annual_cost"](tea, ppl):.1f} {unit}.')
-        print(f'Capital: {func["get_annual_CAPEX"](tea, ppl):.1f} {unit}.')
-        print(f'Operating: {func["get_annual_OPEX"](tea, ppl):.1f} {unit}.')
-
-        unit = f'{GWP.unit}/cap/yr'
-        print(f'\nNet emission: {func["get_annual_GWP"](lca, ppl):.1f} {unit}.')
-        print(f'Construction: {func["get_constr_GWP"](lca, ppl):.1f} {unit}.')
-        print(f'Transportation: {func["get_trans_GWP"](lca, ppl):.1f} {unit}.')
-        print(f'Direct emission: {func["get_direct_emission_GWP"](lca, ppl):.1f} {unit}.')
-        print(f'Offset: {func["get_offset_GWP"](lca, ppl):.1f} {unit}.')
-        print(f'Other: {func["get_other_GWP"](lca, ppl):.1} {unit}.\n')
-
         for i in ('COD', 'N', 'P', 'K'):
             print(f'Total {i} recovery is {func[f"get_tot_{i}_recovery"](sys, i):.1%}, '
                   f'{func[f"get_liq_{i}_recovery"](sys, i):.1%} in liquid, '
                   f'{func[f"get_sol_{i}_recovery"](sys, i):.1%} in solid, '
                   f'{func[f"get_gas_{i}_recovery"](sys, i):.1%} in gas.')
+        print('\n')
+
+        tea = sys_dct['TEA'][sys.ID]
+        tea.show()
+
+        unit = f'{currency}/cap/yr'
+        print(f'\nNet cost: {func["get_annual_net_cost"](tea, ppl):.1f} {unit}.')
+        print(f'Capital: {func["get_annual_CAPEX"](tea, ppl):.1f} {unit}.')
+        print(f'Operating: {func["get_annual_OPEX"](tea, ppl):.1f} {unit}.')
+        print(f'Sales: {func["get_annual_sales"](tea, ppl):.1f} {unit}.')
+        print('\n')
+
+        lca = sys_dct['LCA'][sys.ID]
+        lca.show()
+        print('\n')
+
+        for ind in lca.indicators:
+            unit = f'{ind.unit}/cap/yr'
+            print(f'\nImpact indicator {ind.ID}:')
+
+            val = lca.total_impacts[ind.ID]/lca.lifetime/ppl
+            print(f'\nNet emission: {val:.1f} {unit}.')
+
+            val = lca.total_construction_impacts[ind.ID]/lca.lifetime/ppl
+            print(f'Construction: {val:.1f} {unit}.')
+
+            val = lca.total_transportation_impacts[ind.ID]/lca.lifetime/ppl
+            print(f'Transportation: {val:.1f} {unit}.')
+
+            val = lca.get_stream_impacts(stream_items=lca.stream_inventory, kind='direct_emission')[ind.ID] \
+                /lca.lifetime/ppl
+            print(f'Direct emission: {val:.1f} {unit}.')
+
+            val = lca.get_stream_impacts(stream_items=lca.stream_inventory, kind='offset')[ind.ID] \
+                /lca.lifetime/ppl
+            print(f'Offset: {val:.1f} {unit}.')
+
+            val = lca.total_other_impacts[ind.ID]/lca.lifetime/ppl
+            print(f'Other: {val:.1} {unit}.\n')
+
 
 def save_all_reports():
     import os
-    path = os.path.dirname(os.path.realpath(__file__))
-    path += '/results'
-    if not os.path.isdir(path):
-        os.path.mkdir(path)
-    del os
+    if not os.path.isdir(results_path):
+        os.path.mkdir(results_path)
     for i in (sysA, sysB, sysC, lcaA, lcaB, lcaC):
         if isinstance(i, bst.System):
             i.simulate()
-            i.save_report(f'{path}/{i.ID}.xlsx')
+            i.save_report(os.path.join(results_path, f'{i.ID}_report.xlsx'))
         else:
-            i.save_report(f'{path}/{i.system.ID}_lca.xlsx')
+            i.save_report(os.path.join(results_path, f'{i.system.ID}_lca.xlsx'))
 
 __all__ = ('sysA', 'sysB', 'sysC', 'teaA', 'teaB', 'teaC', 'lcaA', 'lcaB', 'lcaC',
            'print_summaries', 'save_all_reports', 'update_lca_data',
